//
// Copyright (c) 2004 - InfoMus Lab - DIST - University of Genova
//
// InfoMus Lab (Laboratorio di Informatica Musicale)
// DIST - University of Genova
//
// http://www.infomus.dist.unige.it
// news://infomus.dist.unige.it
// mailto:staff@infomus.dist.unige.it
//
// Developer: Gualtiero Volpe
// mailto:volpe@infomus.dist.unige.it
//
// Modified: Oct 25 2006 by Trey Harrison
// email:trey@harrisondigitalmedia.com
//
// Last modified: Oct 01 2018 by Menno Vink
// email:menno@resolume.com

#include "PluginManager.h"
#include "PluginBase.h"

#include <stdlib.h>
#include <memory.h>
#include <algorithm>

namespace ffgl
{

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PluginManager constructor and destructor
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PluginManager::PluginManager()
{
	m_iMinInputs = 0;
	m_iMaxInputs = 0;
	m_timeSupported = true;

	m_NParams = 0;
	m_pFirst = NULL;
	m_pLast = NULL;
}

// 2019-07-22 ffgl@subpixels.com - use enum ParameterType
PluginManager::~PluginManager()
{
	if( m_pFirst != NULL )
	{
		ParamInfo* pCurr = m_pFirst;
		ParamInfo* pNext = m_pFirst;

		while( pCurr != NULL )
		{
			pNext = pCurr->pNext;

			if( ( pCurr->type == ParameterType::FF_TYPE_TEXT ) &&
				( pCurr->StrDefaultValue != NULL ) )
			{
				free( pCurr->StrDefaultValue );
			}

			delete pCurr;
			pCurr = pNext;
		}
	}

	m_pFirst = NULL;
	m_pLast = NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PluginManager methods
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned int PluginManager::GetMinInputs() const
{
	return m_iMinInputs;
}

unsigned int PluginManager::GetMaxInputs() const
{
	return m_iMaxInputs;
}

unsigned int PluginManager::GetNumParams() const
{
	return m_NParams;
}

void PluginManager::SetMinInputs( unsigned int iMinInputs )
{
	m_iMinInputs = iMinInputs;
}

void PluginManager::SetMaxInputs( unsigned int iMaxInputs )
{
	m_iMaxInputs = iMaxInputs;
}

// 2019-07-22 ffgl@subpixels.com - use enums ParameterType, ParameterUsage
void PluginManager::SetParamInfo( unsigned int pIndex, const char* pchName, ParameterType type, float fDefaultValue )
{
	ParamInfo* pInfo = new ParamInfo;
	pInfo->ID = pIndex;
	
	pInfo->elements.resize( 1 );
	pInfo->usage = ParameterUsage::FF_USAGE_STANDARD;

	std::string stringValue = pchName;
	memset(pInfo->Name, 0, sizeof(pInfo->Name));
	memcpy(pInfo->Name, stringValue.c_str(), std::min(sizeof(pInfo->Name), stringValue.length()));

	pInfo->type = type;
	if (type == ParameterType::FF_TYPE_STANDARD)
	{
		if( fDefaultValue > 1.0 )
			fDefaultValue = 1.0;
		if( fDefaultValue < 0.0 )
			fDefaultValue = 0.0;
	}

	pInfo->DefaultValue = fDefaultValue;
	pInfo->StrDefaultValue = NULL;
	pInfo->pNext = NULL;
	if( m_pFirst == NULL )
		m_pFirst = pInfo;
	if( m_pLast != NULL )
		m_pLast->pNext = pInfo;
	m_pLast = pInfo;
	m_NParams++;
}

// 2019-07-22 ffgl@subpixels.com - use enums ParameterType, ParameterUsage
void PluginManager::SetBufferParamInfo( unsigned int pIndex, const char* pchName, unsigned int numElements, ParameterUsage usage )
{
	ParamInfo* pInfo = new ParamInfo;
	pInfo->ID = pIndex;

	pInfo->elements.resize( numElements );
	pInfo->usage = usage;

	std::string stringValue = pchName;
	memset(pInfo->Name, 0, sizeof(pInfo->Name));
	memcpy(pInfo->Name, stringValue.c_str(), std::min(sizeof(pInfo->Name), stringValue.length()));

	pInfo->type = ParameterType::FF_TYPE_BUFFER;

	pInfo->DefaultValue = 0.0f;
	pInfo->StrDefaultValue = NULL;
	pInfo->pNext = NULL;
	if( m_pFirst == NULL )
		m_pFirst = pInfo;
	if( m_pLast != NULL )
		m_pLast->pNext = pInfo;
	m_pLast = pInfo;
	m_NParams++;
}

// 2019-07-22 ffgl@subpixels.com - use enums ParameterType, ParameterUsage
void PluginManager::SetOptionParamInfo( unsigned int pIndex, const char* pchName, unsigned int numElements, float defaultValue )
{
	ParamInfo* pInfo = new ParamInfo;
	pInfo->ID = pIndex;

	pInfo->elements.resize( numElements );
	pInfo->usage = ParameterUsage::FF_USAGE_STANDARD;

	std::string stringValue = pchName;
	memset(pInfo->Name, 0, sizeof(pInfo->Name));
	memcpy(pInfo->Name, stringValue.c_str(), std::min(sizeof(pInfo->Name), stringValue.length()));

	pInfo->type = ParameterType::FF_TYPE_OPTION;

	pInfo->DefaultValue = defaultValue;
	pInfo->StrDefaultValue = NULL;
	pInfo->pNext = NULL;
	if( m_pFirst == NULL )
		m_pFirst = pInfo;
	if( m_pLast != NULL )
		m_pLast->pNext = pInfo;
	m_pLast = pInfo;
	m_NParams++;
}

void PluginManager::SetParamElementInfo( unsigned int paramID, unsigned int elementIndex, const char* elementName, float elementValue )
{
	ParamInfo* paramInfo = FindParamInfo( paramID );
	if( paramInfo == nullptr )
		return;

	if( elementIndex >= paramInfo->elements.size() )
		return;
	
	paramInfo->elements[ elementIndex ].name = elementName;
	paramInfo->elements[ elementIndex ].value = elementValue;
}

// 2019-07-22 ffgl@subpixels.com - use enums ParameterType
void PluginManager::SetParamInfo( unsigned int pIndex, const char* pchName, ParameterType type, bool bDefaultValue )
{
	ParamInfo* pInfo = new ParamInfo;
	pInfo->ID = pIndex;

	std::string stringValue = pchName;
	memset(pInfo->Name, 0, sizeof(pInfo->Name));
	memcpy(pInfo->Name, stringValue.c_str(), std::min(sizeof(pInfo->Name), stringValue.length()));

	pInfo->type = type;
	pInfo->DefaultValue = bDefaultValue ? 1.0f : 0.0f;
	pInfo->StrDefaultValue = NULL;
	pInfo->pNext = NULL;
	if( m_pFirst == NULL )
		m_pFirst = pInfo;
	if( m_pLast != NULL )
		m_pLast->pNext = pInfo;
	m_pLast = pInfo;
	m_NParams++;
}

// 2019-07-22 ffgl@subpixels.com - use enums ParameterType, ParameterUsage
void PluginManager::SetParamInfo( unsigned int dwIndex, const char* pchName, ParameterType type, const char* pchDefaultValue )
{
	ParamInfo* pInfo = new ParamInfo;
	pInfo->ID = dwIndex;

	pInfo->elements.resize( 1 );
	pInfo->usage = ParameterUsage::FF_USAGE_STANDARD;

	std::string stringValue = pchName;
	memset(pInfo->Name, 0, sizeof(pInfo->Name));
	memcpy(pInfo->Name, stringValue.c_str(), std::min(sizeof(pInfo->Name), stringValue.length()));

	pInfo->type = type;
	pInfo->DefaultValue = 0;
	pInfo->StrDefaultValue = _strdup( pchDefaultValue );
	pInfo->pNext = NULL;
	if( m_pFirst == NULL )
		m_pFirst = pInfo;
	if( m_pLast != NULL )
		m_pLast->pNext = pInfo;
	m_pLast = pInfo;
	m_NParams++;
}

void PluginManager::SetTimeSupported( bool supported )
{
	m_timeSupported = supported;
}

void PluginManager::SetParamRange(unsigned int index, float min, float max)
{
	ParamInfo* paramInfo = FindParamInfo(index);
	if (paramInfo)
		paramInfo->range = { min, max };
}

// 2019-07-12 ffgl@subpixels.com - Make method const
RangeStruct PluginManager::GetParamRange( unsigned int index ) const
{
	RangeStruct result = { 0, 1 };
	const ParamInfo* paramInfo = GetParamInfo(index);
	if (paramInfo)
		result = paramInfo->range;
	return result;
}

// 2019-07-12 ffgl@subpixels.com - Make method const and return const
const char* PluginManager::GetParamName( unsigned int dwIndex ) const
{
	const ParamInfo* paramInfo = GetParamInfo( dwIndex );
	if( paramInfo == nullptr )
		return nullptr;

	return paramInfo->Name;
}

// 2019-07-22 ffgl@subpixels.com - use enums ParameterType
ParameterType PluginManager::GetParamType( unsigned int dwIndex ) const
{
	const ParamInfo* paramInfo = GetParamInfo( dwIndex );
	if( paramInfo == nullptr )
		return ParameterType::FF_FAIL;

	return paramInfo->type;
}

FFResult PluginManager::GetNumParamElements( unsigned int dwIndex ) const
{
	const ParamInfo* paramInfo = GetParamInfo( dwIndex );
	if( paramInfo == nullptr )
		return FF_FAIL;

	return (unsigned int)paramInfo->elements.size();
}

// 2019-07-12 ffgl@subpixels.com - Make method const and return const
const char* PluginManager::GetParamElementName( unsigned int dwIndex, unsigned int elIndex ) const
{
	const ParamInfo* paramInfo = GetParamInfo( dwIndex );
	if( paramInfo == nullptr )
		return nullptr;

	if( elIndex >= paramInfo->elements.size() )
		return nullptr;

	return paramInfo->elements[ elIndex ].name.c_str();
}

// 2019-07-12 ffgl@subpixels.com - use FFMixed.FloatValue member instead of pointer recasting
FFMixed PluginManager::GetParamElementDefault( unsigned int dwIndex, unsigned int elIndex ) const
{
	FFMixed result;
	result.UIntValue = FF_FAIL;

	const ParamInfo* paramInfo = GetParamInfo( dwIndex );
	if( paramInfo == nullptr )
		return result;

	if( elIndex >= paramInfo->elements.size() )
		return result;

	result.FloatValue = paramInfo->elements[ elIndex ].value;
	return result;
}

FFResult PluginManager::SetParamElementValue( unsigned int dwIndex, unsigned int elIndex, float newValue )
{
	ParamInfo* paramInfo = FindParamInfo( dwIndex );
	if( paramInfo == nullptr )
		return FF_FAIL;

	if( elIndex >= paramInfo->elements.size() )
		return FF_FAIL;

	paramInfo->elements[ elIndex ].value = newValue;
	return FF_SUCCESS;
}

ParameterUsage PluginManager::GetParamUsage( unsigned int dwIndex ) const
{
	const ParamInfo* paramInfo = GetParamInfo( dwIndex );
	if( paramInfo == nullptr )
		return ParameterUsage::FF_FAIL;

	return paramInfo->usage;
}

// 2019-07-11 ffgl@subpixels.com - use specific FFMixed union members instead of pointer recasting
// 2019-07-22 ffgl@subpixels.com - use enum ParameterType
FFMixed PluginManager::GetParamDefault( unsigned int dwIndex ) const
{
	FFMixed result;
	result.UIntValue = FF_FAIL;

	const ParamInfo* paramInfo = GetParamInfo( dwIndex );
	if( paramInfo == nullptr )
		return result;

	if( paramInfo->type == ParameterType::FF_TYPE_TEXT )
		result.StrValue = paramInfo->StrDefaultValue;
	else
		result.FloatValue = paramInfo->DefaultValue;
	
	return result;
}

bool PluginManager::GetTimeSupported() const
{
	return m_timeSupported;
}

// Find ParamInfo (for editing)
PluginManager::ParamInfo* PluginManager::FindParamInfo( unsigned int ID )
{
	for( ParamInfo* pCurr = m_pFirst; pCurr != nullptr; pCurr = pCurr->pNext )
	{
		if( pCurr->ID == ID )
			return pCurr;
	}
	return nullptr;
}

// Get ParamInfo (read only)
// 2019-07-12 ffgl@subpixels.com - Renamed from FindParamInfo to make distinction (read only) clear
const PluginManager::ParamInfo* PluginManager::GetParamInfo( unsigned int ID ) const
{
	for( const ParamInfo* pCurr = m_pFirst; pCurr != nullptr; pCurr = pCurr->pNext )
	{
		if( pCurr->ID == ID )
			return pCurr;
	}
	return nullptr;
}

}//End namespace ffgl
